# make sure you use grpc version 1.39.0 or later,
# because of https://github.com/grpc/grpc/issues/15880 that affected earlier versions
import grpc
import hello_pb2_grpc
import hello_pb2
from locust import events, User, task
from locust.exception import LocustError
from locust.user.task import LOCUST_STATE_STOPPING
from hello_server import start_server
import gevent
import time

# patch grpc so that it uses gevent instead of asyncio
import grpc.experimental.gevent as grpc_gevent

grpc_gevent.init_gevent()


@events.init.add_listener
def run_grpc_server(environment, **_kwargs):
    # Start the dummy server. This is not something you would do in a real test.
    gevent.spawn(start_server)


class GrpcClient:
    def __init__(self, environment, stub):
        self.env = environment
        self._stub_class = stub.__class__
        self._stub = stub

    def __getattr__(self, name):
        func = self._stub_class.__getattribute__(self._stub, name)

        def wrapper(*args, **kwargs):
            request_meta = {
                "request_type": "grpc",
                "name": name,
                "start_time": time.time(),
                "response_length": 0,
                "exception": None,
                "context": None,
                "response": None,
            }
            start_perf_counter = time.perf_counter()
            try:
                request_meta["response"] = func(*args, **kwargs)
                request_meta["response_length"] = len(request_meta["response"].message)
            except grpc.RpcError as e:
                request_meta["exception"] = e
            request_meta["response_time"] = (time.perf_counter() - start_perf_counter) * 1000
            self.env.events.request.fire(**request_meta)
            return request_meta["response"]

        return wrapper


class GrpcUser(User):
    abstract = True

    stub_class = None

    def __init__(self, environment):
        super().__init__(environment)
        for attr_value, attr_name in ((self.host, "host"), (self.stub_class, "stub_class")):
            if attr_value is None:
                raise LocustError(f"You must specify the {attr_name}.")
        self._channel = grpc.insecure_channel(self.host)
        self._channel_closed = False
        stub = self.stub_class(self._channel)
        self.client = GrpcClient(environment, stub)


class HelloGrpcUser(GrpcUser):
    host = "localhost:50051"
    stub_class = hello_pb2_grpc.HelloServiceStub

    @task
    def sayHello(self):
        if not self._channel_closed:
            self.client.SayHello(hello_pb2.HelloRequest(name="Test"))
        time.sleep(1)



from locust import User, task, constant
from locust.user.sequential_taskset import SequentialTaskSet
from locust.exception import RescheduleTask
from .testcases import LocustTestCase


class TestTaskSet(LocustTestCase):
    def setUp(self):
        super().setUp()

        class MyUser(User):
            host = "127.0.0.1"

        self.locust = MyUser(self.environment)

    def test_task_sequence_with_list(self):
        log = []

        def t1(ts):
            log.append(1)

        def t2(ts):
            log.append(2)

        def t3(ts):
            log.append(3)
            ts.interrupt(reschedule=False)

        class MyTaskSequence(SequentialTaskSet):
            tasks = [t1, t2, t3]

        l = MyTaskSequence(self.locust)
        self.assertRaises(RescheduleTask, lambda: l.run())
        self.assertEqual([1, 2, 3], log)

    def test_task_sequence_with_methods(self):
        log = []

        class MyTaskSequence(SequentialTaskSet):
            @task
            def t1(self):
                log.append(1)

            @task
            def t2(self):
                log.append(2)

            @task(1)
            def t3(self):
                log.append(3)
                self.interrupt(reschedule=False)

        l = MyTaskSequence(self.locust)
        self.assertRaises(RescheduleTask, lambda: l.run())
        self.assertEqual([1, 2, 3], log)

    def test_task_sequence_with_methods_and_list(self):
        log = []

        def func_t1(ts):
            log.append(101)

        def func_t2(ts):
            log.append(102)

        class MyTaskSequence(SequentialTaskSet):
            @task
            def t1(self):
                log.append(1)

            @task
            def t2(self):
                log.append(2)

            tasks = [func_t1, func_t2]

            @task(1)
            def t3(self):
                log.append(3)
                self.interrupt(reschedule=False)

        l = MyTaskSequence(self.locust)
        self.assertRaises(RescheduleTask, lambda: l.run())
        self.assertEqual([1, 2, 101, 102, 3], log)

    def test_task_sequence_with_inheritance(self):
        log = []

        class TS1(SequentialTaskSet):
            @task
            def t1(self):
                log.append(1)

            tasks = [lambda ts: log.append(30)]

        class TS2(TS1):
            tasks = [lambda ts: log.append(20)]

            @task
            def t2(self):
                log.append(2)

        class TS3(TS2):
            @task
            def t3(self):
                log.append(3)
                self.interrupt(reschedule=False)

        l = TS3(self.locust)
        self.assertRaises(RescheduleTask, lambda: l.run())
        self.assertEqual([1, 30, 20, 2, 3], log)

    def test_task_sequence_multiple_iterations(self):
        log = []

        class TS(SequentialTaskSet):
            iteration_count = 0

            @task
            def t1(self):
                log.append(1)

            @task
            def t2(self):
                log.append(2)

            @task(1)
            def t3(self):
                log.append(3)
                self.iteration_count += 1
                if self.iteration_count == 3:
                    self.interrupt(reschedule=False)

        l = TS(self.locust)
        self.assertRaises(RescheduleTask, lambda: l.run())
        self.assertEqual([1, 2, 3, 1, 2, 3, 1, 2, 3], log)



from locust import HttpUser, TaskSet, task, between


class WebsiteUser(HttpUser):
    """
    Example of the ability of inline nested TaskSet classes
    """

    host = "http://127.0.0.1:8089"
    wait_time = between(2, 5)

    @task
    class TopLevelTaskSet(TaskSet):
        @task
        class IndexTaskSet(TaskSet):
            @task(10)
            def index(self):
                self.client.get("/")

            @task(1)
            def stop(self):
                self.interrupt()

        @task
        def stats(self):
            self.client.get("/stats/requests")



import time
from locust import HttpUser, task, between


class QuickstartUser(HttpUser):
    wait_time = between(1, 2)

    @task
    def hello_world(self):
        self.client.get("/hello")
        self.client.get("/world")

    @task(3)
    def view_item(self):
        for item_id in range(10):
            self.client.get(f"/item?id={item_id}", name="/item")
            time.sleep(1)

    def on_start(self):
        self.client.post("/login", json={"username": "foo", "password": "bar"})



import os

if os.getenv("LOCUST_PLAYWRIGHT", False):
    # This is a hack to make Playwright testing possible. If trio is not imported before gevent's monkey patching,
    # it raises "NotImplementedError: unsupported platform"
    # Playwright is used by PlaywrightUser, see https://github.com/SvenskaSpel/locust-plugins/blob/master/examples/playwright_ex.py
    import trio

from gevent import monkey

monkey.patch_all()

from ._version import version as __version__
from .user.sequential_taskset import SequentialTaskSet
from .user import wait_time
from .user.task import task, tag, TaskSet
from .user.users import HttpUser, User
from .contrib.fasthttp import FastHttpUser
from .user.wait_time import between, constant, constant_pacing, constant_throughput
from .shape import LoadTestShape
from .debug import run_single_user

from .event import Events

events = Events()

__all__ = (
    "SequentialTaskSet",
    "wait_time",
    "task",
    "tag",
    "TaskSet",
    "HttpUser",
    "FastHttpUser",
    "User",
    "between",
    "constant",
    "constant_pacing",
    "constant_throughput",
    "events",
    "LoadTestShape",
)

# Used for raising a DeprecationWarning if old Locust/HttpLocust is used
from .util.deprecation import DeprecatedLocustClass as Locust
from .util.deprecation import DeprecatedHttpLocustClass as HttpLocust



import os

from locust import HttpUser, TaskSet, task, between
from locust.clients import HttpSession


class MultipleHostsUser(HttpUser):
    abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_client = HttpSession(
            base_url=os.environ["API_HOST"], request_event=self.client.request_event, user=self
        )


class UserTasks(TaskSet):
    # but it might be convenient to use the @task decorator
    @task
    def index(self):
        self.user.client.get("/")

    @task
    def index_other_host(self):
        self.user.api_client.get("/stats/requests")


class WebsiteUser(MultipleHostsUser):
    """
    User class that does requests to the locust web server running on localhost
    """

    host = "http://127.0.0.1:8089"
    wait_time = between(2, 5)
    tasks = [UserTasks]



from jinja2 import Environment, FileSystemLoader
import os
import pathlib
import datetime
from itertools import chain
from .stats import sort_stats
from .user.inspectuser import get_ratio
from html import escape
from json import dumps
from .runners import MasterRunner, STATE_STOPPED, STATE_STOPPING


def render_template(file, **kwargs):
    templates_path = os.path.join(pathlib.Path(__file__).parent.absolute(), "templates")
    env = Environment(loader=FileSystemLoader(templates_path), extensions=["jinja2.ext.do"])
    template = env.get_template(file)
    return template.render(**kwargs)


def get_html_report(environment, show_download_link=True):
    stats = environment.runner.stats

    start_ts = stats.start_time
    start_time = datetime.datetime.utcfromtimestamp(start_ts).strftime("%Y-%m-%d %H:%M:%S")

    end_ts = stats.last_request_timestamp
    if end_ts:
        end_time = datetime.datetime.utcfromtimestamp(end_ts).strftime("%Y-%m-%d %H:%M:%S")
    else:
        end_time = start_time

    host = None
    if environment.host:
        host = environment.host
    elif environment.runner.user_classes:
        all_hosts = {l.host for l in environment.runner.user_classes}
        if len(all_hosts) == 1:
            host = list(all_hosts)[0]

    requests_statistics = list(chain(sort_stats(stats.entries), [stats.total]))
    failures_statistics = sort_stats(stats.errors)
    exceptions_statistics = [
        {**exc, "nodes": ", ".join(exc["nodes"])} for exc in environment.runner.exceptions.values()
    ]

    history = stats.history

    static_js = []
    js_files = ["jquery-1.11.3.min.js", "echarts.common.min.js", "vintage.js", "chart.js", "tasks.js"]
    for js_file in js_files:
        path = os.path.join(os.path.dirname(__file__), "static", js_file)
        static_js.append("// " + js_file)
        with open(path, encoding="utf8") as f:
            static_js.append(f.read())
        static_js.extend(["", ""])

    static_css = []
    css_files = ["tables.css"]
    for css_file in css_files:
        path = os.path.join(os.path.dirname(__file__), "static", "css", css_file)
        static_css.append("/* " + css_file + " */")
        with open(path, encoding="utf8") as f:
            static_css.append(f.read())
        static_css.extend(["", ""])

    is_distributed = isinstance(environment.runner, MasterRunner)
    user_spawned = (
        environment.runner.reported_user_classes_count if is_distributed else environment.runner.user_classes_count
    )

    if environment.runner.state in [STATE_STOPPED, STATE_STOPPING]:
        user_spawned = environment.runner.final_user_classes_count

    task_data = {
        "per_class": get_ratio(environment.user_classes, user_spawned, False),
        "total": get_ratio(environment.user_classes, user_spawned, True),
    }

    res = render_template(
        "report.html",
        int=int,
        round=round,
        escape=escape,
        str=str,
        requests_statistics=requests_statistics,
        failures_statistics=failures_statistics,
        exceptions_statistics=exceptions_statistics,
        start_time=start_time,
        end_time=end_time,
        host=host,
        history=history,
        static_js="\n".join(static_js),
        static_css="\n".join(static_css),
        show_download_link=show_download_link,
        locustfile=environment.locustfile,
        tasks=dumps(task_data),
    )

    return res



import functools
from time import time


def memoize(timeout, dynamic_timeout=False):
    """
    Memoization decorator with support for timeout.

    If dynamic_timeout is set, the cache timeout is doubled if the cached function
    takes longer time to run than the timeout time
    """
    cache = {"timeout": timeout}

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start = time()
            if (not "time" in cache) or (start - cache["time"] > cache["timeout"]):
                # cache miss
                cache["result"] = func(*args, **kwargs)
                cache["time"] = time()
                if dynamic_timeout and cache["time"] - start > cache["timeout"]:
                    cache["timeout"] *= 2
            return cache["result"]

        def clear_cache():
            if "time" in cache:
                del cache["time"]
            if "result" in cache:
                del cache["result"]

        wrapper.clear_cache = clear_cache
        return wrapper

    return decorator



import gevent
from locust import HttpUser, task, between
from locust.env import Environment
from locust.stats import stats_printer, stats_history
from locust.log import setup_logging

setup_logging("INFO", None)


class User(HttpUser):
    wait_time = between(1, 3)
    host = "https://docs.locust.io"

    @task
    def my_task(self):
        self.client.get("/")

    @task
    def task_404(self):
        self.client.get("/non-existing-path")


# setup Environment and Runner
env = Environment(user_classes=[User])
env.create_local_runner()

# start a WebUI instance
env.create_web_ui("127.0.0.1", 8089)

# start a greenlet that periodically outputs the current stats
gevent.spawn(stats_printer(env.stats))

# start a greenlet that save current stats to history
gevent.spawn(stats_history, env.runner)

# start the test
env.runner.start(1, spawn_rate=10)

# in 60 seconds stop the runner
gevent.spawn_later(60, lambda: env.runner.quit())

# wait for the greenlets
env.runner.greenlet.join()

# stop the web server for good measures
env.web_ui.stop()



from locust import HttpUser, task, events, between
from locust.runners import MasterRunner, WorkerRunner

usernames = []


def setup_test_users(environment, msg, **kwargs):
    # Fired when the worker receives a message of type 'test_users'
    usernames.extend(map(lambda u: u["name"], msg.data))
    environment.runner.send_message("acknowledge_users", f"Thanks for the {len(msg.data)} users!")


def on_acknowledge(msg, **kwargs):
    # Fired when the master receives a message of type 'acknowledge_users'
    print(msg.data)


@events.init.add_listener
def on_locust_init(environment, **_kwargs):
    if not isinstance(environment.runner, MasterRunner):
        environment.runner.register_message("test_users", setup_test_users)
    if not isinstance(environment.runner, WorkerRunner):
        environment.runner.register_message("acknowledge_users", on_acknowledge)


@events.test_start.add_listener
def on_test_start(environment, **_kwargs):
    # When the test is started, evenly divides list between
    # worker nodes to ensure unique data across threads
    if not isinstance(environment.runner, WorkerRunner):
        users = []
        for i in range(environment.runner.target_user_count):
            users.append({"name": f"User{i}"})

        worker_count = environment.runner.worker_count
        chunk_size = int(len(users) / worker_count)

        for i, worker in enumerate(environment.runner.clients):
            start_index = i * chunk_size

            if i + 1 < worker_count:
                end_index = start_index + chunk_size
            else:
                end_index = len(users)

            data = users[start_index:end_index]
            environment.runner.send_message("test_users", data, worker)


class WebsiteUser(HttpUser):
    host = "http://127.0.0.1:8089"
    wait_time = between(2, 5)

    def __init__(self, parent):
        self.username = usernames.pop()
        super().__init__(parent)

    @task
    def task(self):
        print(self.username)
